import sys
import kgd.constants as cnst


def main():
    try:
        from kgd.core import main
        exit_status = main()
        if exit_status == 2:
            print(cnst.NO_BINS)
        elif exit_status == 3:
            print(cnst.TOO_MANY_FAILS)
        elif exit_status == 4:
            print(cnst.SRV_IS_DOWN)
    except KeyboardInterrupt:
        from kgd.constants import ExitStatus
        exit_status = ExitStatus.ERROR_CTRL_C

    sys.exit(exit_status)


if __name__ == "__main__":
    main()



import argparse

import kgd.validators as validators


def parse_args():
    """ Parse CLI arguments and return validated values """
    parser = argparse.ArgumentParser(
        description="Tool for retrieving information on tax payments by Kazakhstan companies",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("token", type=str,
                        help="KGD API secret token")

    parser.add_argument("fpath", type=validators.check_fpath,
                        help="input file with BINs")

    parser.add_argument("date_range", type=validators.check_date_range,
                        help="date range for which we retrieve data")

    parser.add_argument("-t", "--timeout", default=3.0,
                        type=validators.check_positive_float,
                        help="server connect timeout")

    parser.add_argument("-f", "--fsize",
                        default=50000000, type=validators.check_positive_int,
                        help="size limit for output file")

    return parser.parse_args()



import os
from collections import deque
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

import attr
import requests
import xmltodict
from box import Box

from kgd.constants import HOST
from kgd.exceptions import KgdTooManyRequests
from kgd.utils import (read_file, requests_retry_session,
                       run_command, append_file)
from kgd.validators import common_corrector, date_corrector


def is_bin_processed(bn, processed_fpath):
    """ Check if BIN has been already processed """
    r = run_command(['grep', bn, processed_fpath])
    if r:
        return True

    return False


def processed_bins_fpath(fpath):
    """ Return path to file with processed BINs """
    return os.path.splitext(fpath)[0] + '.prsd'


@attr.s
class PaymentData:
    """ Wrap dict and validate/convert values of each field"""
    bin = attr.ib(default='')
    taxorgcode = attr.ib(converter=common_corrector, default='')
    nametaxru = attr.ib(converter=common_corrector, default='')
    nametaxkz = attr.ib(converter=common_corrector, default='')
    kbk = attr.ib(converter=common_corrector, default='')
    kbknameru = attr.ib(converter=common_corrector, default='')
    kbknamekz = attr.ib(converter=common_corrector, default='')
    paynum = attr.ib(converter=common_corrector, default='')
    paytype = attr.ib(converter=common_corrector, default='')
    entrytype = attr.ib(converter=common_corrector, default='')
    receiptdate = attr.ib(converter=date_corrector, default='')
    writeoffdate = attr.ib(converter=date_corrector, default='')
    summa = attr.ib(converter=common_corrector, default='')


class TaxPaymentParser:
    # TODO request.xml moved into kgd folder
    request_template = read_file(
        os.path.join(os.path.abspath(
            os.path.join(os.path.dirname(__file__))), 'request.xml'
        )
    )

    url_template = "https://{}/proxy2/culs_payments?token={}"
    headers = {'user-agent': 'Apache-HttpClient/4.1.1 (java 1.5)',
               'content-type': 'text/xml'}
    # 429 - too many requests
    status_forcelist = (429, 500, 502, 504)

    def __init__(self, fpath, fsize):
        self._output_files = []
        self._add_output_files(fpath, fsize)
        self._failed_bins = deque([])
        self._failed_bins_count = 0
        self.session = requests.Session()
        self.session.headers.update(self.headers)

    def _curr_output_file(self):
        return self._output_files[len(self._output_files)-1]

    @property
    def fails(self):
        return self._failed_bins

    def add_fail(self, bn):
        self._failed_bins.append(bn)

    def pop_failed(self):
        return self._failed_bins.popleft()

    def put_failed(self, bn):
        self._failed_bins.append(bn)

    @property
    def output_file(self):
        return self._curr_output_file()

    def _add_output_files(self, fpath, fsize):
        """
        Gather all existed output files paths and
        if it's needed add new
        """
        self._output_files = []
        base = os.path.join(os.path.dirname(fpath),
                            os.path.splitext(os.path.basename(fpath))[0])

        output_path = f'{base}_out.csv'

        suffix = 2
        while os.path.exists(output_path):
            self._output_files.append(output_path)
            if os.path.getsize(output_path) < fsize:
                return
            output_path = f'{base}_out_{suffix}.csv'
            suffix += 1

        open(output_path, 'a').close()
        self._output_files.append(output_path)

    def process(self, bn, token=None,
                fpath=None, fsize=None, date_range=None, hook=None):
        def csv_payment_row(p_dict):
            values = []
            #
            dct = {k.lower(): v.lower() for k, v in p_dict.items()}
            payment_data = PaymentData(**dct)
            return attr.astuple(payment_data)

        # size limit for output csv file
        if os.path.getsize(self._curr_output_file()) >= fsize:
            self._add_output_files(fpath, fsize)

        payments = []

        # Box for payments
        d = None

        request = self.request_template.format(bn, *date_range)

        url = self.url_template.format(HOST, token)
        response = requests.post(url, request, headers=self.headers, verify=False)

        if response.status_code != 200:
            response.raise_for_status()

        if response.text:
            # convert xml to json with OOP features
            d = Box(xmltodict.parse(response.text)).answer

        else:
            raise KgdTooManyRequests("Failed to process {}".format(bn))

        # check if we got some logical errors in response
        if 'err' in d:
            return

        # it might be just one payment
        payments = d.payment if isinstance(d.payment, list) else [d.payment]

        rows = []

        # get only values and put them to list
        # according order how fields going in PaymentData class
        # build csv rows and write them to current output file
        for p in payments:
            p.bin = bn
            append_file(self.output_file, ';'.join(csv_payment_row(p)))





from enum import IntEnum

HOST = "data.egov.kz"

class ExitStatus(IntEnum):
    """Program exit code constants."""
    SUCCESS = 0
    ERROR = 1
    ERROR_NO_BINS = 2
    ERROR_TOO_MANY_FAILS = 3
    ERROR_SERVER_IS_DOWN = 4

    # 128+2 SIGINT <http://www.tldp.org/LDP/abs/html/exitcodes.html>
    ERROR_CTRL_C = 130


PROCESSED_NOT_EXISTS_MESSAGE = 'There are no processed BINs.'+'\n' +\
                               'Loading BINs to start retrieving data ... '
PROCESSED_EXISTS_MESSAGE = 'There are processed BINs.'+'\n' +\
                           'Loading BINs to continue retrieving data ...'

NO_BINS = 'No BINs to parse'
TOO_MANY_FAILS = 'Too many fails. Exiting...'
SRV_IS_DOWN = 'Address is unreachable. Exiting...'



import socket
import subprocess as subp
from urllib3 import Retry

import requests
from requests.adapters import HTTPAdapter


def load_lines(fpath):
    """ Return rows of file as list"""
    with open(fpath, "r", encoding="utf-8") as f:
        lines = [b.rstrip() for b in f.readlines()]

    return lines


def read_file(fpath):
    """ Return all rows of file as string"""
    with open(fpath, 'r', encoding="utf8") as f:
        data = f.read()

    return data


def append_file(fpath, data):
    """ Add new line to file"""
    with open(fpath, 'a+', encoding="utf8") as f:
        f.write(data + '\n')


def requests_retry_session(retries, backoff,
                           status_forcelist, session=None):
    """ Make request with timeout and retries """
    session = session or requests.Session()
    retry = Retry(total=retries,read=retries, connect=retries,
                  backoff_factor=backoff, status_forcelist=status_forcelist,
                  method_whitelist=frozenset(['GET', 'POST'])
    )

    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    return session


def run_command(args, encoding="utf-8", **kwargs):
    p = subp.Popen(args, stdout=subp.PIPE, stderr=subp.PIPE, **kwargs)
    result, err = p.communicate()
    if p.returncode > 1:
        raise IOError(err)

    if p.returncode == 1 and result:
        return result.decode(encoding).strip().split()[0]

    return None


def is_server_up(address, port=443):
    is_up = True
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        result = sock.connect_ex((address, int(port)))
    except Exception:
        is_up = False

    return is_up



import argparse
import datetime
import ipaddress
import os


def check_fpath(value):
    if not os.path.exists(value):
        raise argparse.ArgumentTypeError(
            f"{value} doesn't exists")
    return value


def check_ip_port(value):
    def fail_address_port_format():
        raise argparse.ArgumentTypeError(
            f"{value} is not a correct address:port pair")

    try:
        addr, port = value.split(':')
        ip = ipaddress.ip_address(addr)
        port = int(port)
        if not 1 <= port <= 65535:
            raise ValueError('Wrong port')
    except ValueError:
        raise fail_address_port_format()

    return addr + ':' + str(port)


def check_positive_float(value):
    def fail():
        raise argparse.ArgumentTypeError(
            f"{value} is not a valid value")
    try:
        fvalue = float(value)
    except ValueError:
        fail()
    if fvalue <= 0:
        fail()
    return fvalue


def check_positive_int(value):
    def fail():
        raise argparse.ArgumentTypeError(
            f"{value} is not a valid value")
    try:
        fvalue = int(value)
    except ValueError:
        fail()
    if fvalue <= 0:
        fail()
    return fvalue


def check_date_range(value):
    def fail():
        raise argparse.ArgumentTypeError(
            f"{value} is not a valid date range")
    try:
        date_range = value.split(':')
        if len(date_range) != 2:
            raise ValueError('Date range error')

        begin_dt = datetime.datetime.strptime(date_range[0], '%Y-%m-%d')
        end_dt = datetime.datetime.strptime(date_range[1], '%Y-%m-%d')
        if begin_dt > end_dt:
            fail()

        fvalue = date_range[0], date_range[1]
    except ValueError:
        fail()

    return fvalue


def common_corrector(value):
    return value.rstrip().replace('"', "'").replace('\n', '')


def date_corrector(value):
    return common_corrector(value).split('+')[0]



import os
import sys
from collections import deque, namedtuple
from enum import Enum
from time import sleep
from typing import NamedTuple

from box import Box
from requests import HTTPError, ConnectionError
from tqdm import tqdm

from kgd.exceptions import KgdTooManyRequests
from kgd.cli import parse_args
from kgd.constants import (PROCESSED_EXISTS_MESSAGE,
                           PROCESSED_NOT_EXISTS_MESSAGE, ExitStatus, HOST)
from kgd.parsing import TaxPaymentParser, processed_bins_fpath
from kgd.utils import load_lines, append_file, is_server_up


class TqdmUpTo(tqdm):
    def update_to(self, filename, bn, fails, reproc=False):
        if reproc:
            caption = 'Reprocessing {} in {} ... Fails: {}'
            cnt = 0
        else:
            caption = 'Processing {} in {} ... Fails: {}'
            cnt = 1

        self.set_description(caption.format(bn, filename, fails))
        self.update(cnt)


def main():
    exit_status = ExitStatus.SUCCESS
    args = parse_args()
    p = Box(vars(args))
    prsd_bins = []

    timeout = p.timeout
    del p["timeout"]

    if not is_server_up(HOST):
        return ExitStatus.ERROR_SERVER_IS_DOWN

    # if file with parsed BINS exists
    # that means we've processed some part of them before
    # also it might be all of them...
    # but with fails(connection, network)
    if not os.path.exists(processed_bins_fpath(p.fpath)):
        print(PROCESSED_NOT_EXISTS_MESSAGE)
        open(processed_bins_fpath(p.fpath), 'a').close()
    else:
        print(PROCESSED_EXISTS_MESSAGE)
        prsd_bins = load_lines(processed_bins_fpath(p.fpath))

    src_bins = load_lines(p.fpath)

    # all BINS minus parsed
    if prsd_bins:
        s = set(src_bins)
        s.difference_update(set(prsd_bins))
        bins = list(s)
    else:
        bins = src_bins

    # no BINS to process
    if not bins:
        return ExitStatus.ERROR_NO_BINS

    failed_bins = deque()
    bins = deque(bins)

    fatal_cnt = 0
    big_timeout = timeout

    total = len(bins)
    parser = TaxPaymentParser(p.fpath, p.fsize)
    with TqdmUpTo(total=total) as pbar:
        is_fail = False
        while bins:
            try:
                # get new BIN if last time we failed
                if is_fail:
                    bn = bins.popleft()
                    reproc = False
                else:
                    # otherwise we can reprocess
                    # old failed BIN
                    if parser.fails:
                        bn = parser.pop_failed()
                        reproc = True
                    else:
                        # no fails, go with new BIN
                        bn = bins.popleft()
                        reproc = False

                fname = os.path.basename(parser.output_file)

                parser.process(bn, **p,
                               hook=pbar.update_to(fname,
                                                   bn,
                                                   len(parser.fails),
                                                   reproc=reproc))
            except (HTTPError, ConnectionError) as e:
                fatal_cnt += 1
                big_timeout += timeout
                if fatal_cnt > 30:
                    print(e)
                    return ExitStatus.ERROR
                sleep(big_timeout)

            except KgdTooManyRequests:
                parser.put_failed(bn)
                is_fail = True
                # exit if it's too many fails
                fails_percent = len(parser.fails) * 100 / total
                if fails_percent > 90:
                    return ExitStatus.ERROR
                sleep(timeout)

            else:
                is_fail = False

                # successfully processed BINs will be excluded next time
                append_file(processed_bins_fpath(p.fpath), bn)

    return ExitStatus.SUCCESS



from setuptools import setup


from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='kgd',
      version='0.1.16',
      description='Tool for retrieving information on tax payments by Kazakhstan companies',
      url='https://github.com/elessarelfstone/kgd',
      author='Dauren Sdykov',
      author_email='elessarelfstone@mail.ru',
      license='MIT',
      packages=['kgd'],
      include_package_data=True,
      python_requires='>=3.6.1',
      setup_requires=[
          'wheel',
      ],
      install_requires=[
          'attrs>=19.3.0',
          'python-box>=3.4.5',
          'requests>=2.22.0',
          'xmltodict>=0.12.0',
          'tqdm>=4.38.0'
      ],
      extras_require={
          'dev': [
              'setuptools>=38.6.0',
              'wheel>=0.31.0',
              'twine>=1.11.0',
          ],
      },
      entry_points={
          'console_scripts': [
              'kgd=kgd.__main__:main',
          ],
      },
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Development Status :: 4 - Beta",
          "Environment :: No Input/Output (Daemon)",
          "Intended Audience :: System Administrators",
          "Natural Language :: English",
          "Topic :: Internet",
      ],
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=True)
